@@ -1,7 +1,6 @@
 import mimetypes
 import sys
 import typing as t
-import warnings
 from collections import defaultdict
 from datetime import datetime
 from datetime import timedelta
@@ -39,6 +38,7 @@ from .urls import url_fix
 from .urls import url_parse
 from .urls import url_unparse
 from .urls import url_unquote
+from .utils import cached_property
 from .utils import get_content_type
 from .wrappers.request import Request
 from .wrappers.response import Response
@@ -56,7 +56,7 @@ def stream_encode_multipart(
     threshold: int = 1024 * 500,
     boundary: t.Optional[str] = None,
     charset: str = "utf-8",
-) -> t.Tuple[t.BinaryIO, int, str]:
+) -> t.Tuple[t.IO[bytes], int, str]:
     """Encode a dict of values (either strings or file descriptors or
     :class:`FileStorage` objects.) into a multipart encoded string stored
     in a file descriptor.
@@ -64,7 +64,7 @@ def stream_encode_multipart(
     if boundary is None:
         boundary = f"---------------WerkzeugFormPart_{time()}{random()}"
 
-    stream: t.BinaryIO = BytesIO()
+    stream: t.IO[bytes] = BytesIO()
     total_length = 0
     on_disk = False
 
@@ -81,7 +81,7 @@ def stream_encode_multipart(
                 if length + total_length <= threshold:
                     stream.write(s)
                 else:
-                    new_stream = t.cast(t.BinaryIO, TemporaryFile("wb+"))
+                    new_stream = t.cast(t.IO[bytes], TemporaryFile("wb+"))
                     new_stream.write(stream.getvalue())  # type: ignore
                     new_stream.write(s)
                     stream = new_stream
@@ -306,6 +306,10 @@ class EnvironBuilder:
         ``Authorization`` header value. A ``(username, password)`` tuple
         is a shortcut for ``Basic`` authorization.
 
+    .. versionchanged:: 2.1
+        ``CONTENT_TYPE`` and ``CONTENT_LENGTH`` are not duplicated as
+        header keys in the environ.
+
     .. versionchanged:: 2.0
         ``REQUEST_URI`` and ``RAW_URI`` is the full raw URI including
         the query string, not only the path.
@@ -322,7 +326,7 @@ class EnvironBuilder:
 
     .. versionadded:: 0.15
         The environ has keys ``REQUEST_URI`` and ``RAW_URI`` containing
-        the path before perecent-decoding. This is not part of the WSGI
+        the path before percent-decoding. This is not part of the WSGI
         PEP, but many WSGI servers include it.
 
     .. versionchanged:: 0.6
@@ -347,7 +351,7 @@ class EnvironBuilder:
 
     _args: t.Optional[MultiDict]
     _query_string: t.Optional[str]
-    _input_stream: t.Optional[t.BinaryIO]
+    _input_stream: t.Optional[t.IO[bytes]]
     _form: t.Optional[MultiDict]
     _files: t.Optional[FileMultiDict]
 
@@ -357,15 +361,17 @@ class EnvironBuilder:
         base_url: t.Optional[str] = None,
         query_string: t.Optional[t.Union[t.Mapping[str, str], str]] = None,
         method: str = "GET",
-        input_stream: t.Optional[t.BinaryIO] = None,
+        input_stream: t.Optional[t.IO[bytes]] = None,
         content_type: t.Optional[str] = None,
         content_length: t.Optional[int] = None,
-        errors_stream: t.Optional[t.TextIO] = None,
+        errors_stream: t.Optional[t.IO[str]] = None,
         multithread: bool = False,
         multiprocess: bool = False,
         run_once: bool = False,
         headers: t.Optional[t.Union[Headers, t.Iterable[t.Tuple[str, str]]]] = None,
-        data: t.Optional[t.Union[t.BinaryIO, str, bytes, t.Mapping[str, t.Any]]] = None,
+        data: t.Optional[
+            t.Union[t.IO[bytes], str, bytes, t.Mapping[str, t.Any]]
+        ] = None,
         environ_base: t.Optional[t.Mapping[str, t.Any]] = None,
         environ_overrides: t.Optional[t.Mapping[str, t.Any]] = None,
         charset: str = "utf-8",
@@ -490,7 +496,7 @@ class EnvironBuilder:
         self,
         key: str,
         value: t.Union[
-            t.BinaryIO, t.Tuple[t.BinaryIO, str], t.Tuple[t.BinaryIO, str, str]
+            t.IO[bytes], t.Tuple[t.IO[bytes], str], t.Tuple[t.IO[bytes], str, str]
         ],
     ) -> None:
         """Called in the EnvironBuilder to add files from the data dict."""
@@ -568,7 +574,7 @@ class EnvironBuilder:
         .. versionadded:: 0.14
         """
 
-        def on_update(d: t.Mapping[str, str]) -> None:
+        def on_update(d: CallbackDict) -> None:
             self.headers["Content-Type"] = dump_options_header(self.mimetype, d)
 
         d = parse_options_header(self.headers.get("content-type", ""))[1]
@@ -638,7 +644,7 @@ class EnvironBuilder:
         self._set_form("_files", value)
 
     @property
-    def input_stream(self) -> t.Optional[t.BinaryIO]:
+    def input_stream(self) -> t.Optional[t.IO[bytes]]:
         """An optional input stream. This is mutually exclusive with
         setting :attr:`form` and :attr:`files`, setting it will clear
         those. Do not provide this if the method is not ``POST`` or
@@ -647,7 +653,7 @@ class EnvironBuilder:
         return self._input_stream
 
     @input_stream.setter
-    def input_stream(self, value: t.Optional[t.BinaryIO]) -> None:
+    def input_stream(self, value: t.Optional[t.IO[bytes]]) -> None:
         self._input_stream = value
         self._form = None
         self._files = None
@@ -786,14 +792,15 @@ class EnvironBuilder:
         )
 
         headers = self.headers.copy()
+        # Don't send these as headers, they're part of the environ.
+        headers.remove("Content-Type")
+        headers.remove("Content-Length")
 
         if content_type is not None:
             result["CONTENT_TYPE"] = content_type
-            headers.set("Content-Type", content_type)
 
         if content_length is not None:
             result["CONTENT_LENGTH"] = str(content_length)
-            headers.set("Content-Length", content_length)
 
         combined_headers = defaultdict(list)
 
@@ -837,6 +844,11 @@ class Client:
     `allow_subdomain_redirects` to `True` as if not no external redirects
     are allowed.
 
+    .. versionchanged:: 2.1
+        Removed deprecated behavior of treating the response as a
+        tuple. All data is available as properties on the returned
+        response object.
+
     .. versionchanged:: 2.0
         ``response_wrapper`` is always a subclass of
         :class:``TestResponse``.
@@ -956,7 +968,9 @@ class Client:
         :meta private:
         """
         scheme, netloc, path, qs, anchor = url_parse(response.location)
-        builder = EnvironBuilder.from_environ(response.request.environ, query_string=qs)
+        builder = EnvironBuilder.from_environ(
+            response.request.environ, path=path, query_string=qs
+        )
 
         to_name_parts = netloc.split(":", 1)[0].split(".")
         from_name_parts = builder.server_name.split(".")
@@ -1011,7 +1025,6 @@ class Client:
     def open(
         self,
         *args: t.Any,
-        as_tuple: bool = False,
         buffered: bool = False,
         follow_redirects: bool = False,
         **kwargs: t.Any,
@@ -1030,6 +1043,9 @@ class Client:
             :attr:`TestResponse.history` lists the intermediate
             responses.
 
+        .. versionchanged:: 2.1
+            Removed the ``as_tuple`` parameter.
+
         .. versionchanged:: 2.0
             ``as_tuple`` is deprecated and will be removed in Werkzeug
             2.1. Use :attr:`TestResponse.request` and
@@ -1075,7 +1091,10 @@ class Client:
         redirects = set()
         history: t.List["TestResponse"] = []
 
-        while follow_redirects and response.status_code in {
+        if not follow_redirects:
+            return response
+
+        while response.status_code in {
             301,
             302,
             303,
@@ -1102,24 +1121,12 @@ class Client:
             history.append(response)
             response = self.resolve_redirect(response, buffered=buffered)
         else:
-            # This is the final request after redirects, or not
-            # following redirects.
+            # This is the final request after redirects.
             response.history = tuple(history)
             # Close the input stream when closing the response, in case
             # the input is an open temporary file.
             response.call_on_close(request.input_stream.close)
-
-        if as_tuple:
-            warnings.warn(
-                "'as_tuple' is deprecated and will be removed in"
-                " Werkzeug 2.1. Access 'response.request.environ'"
-                " instead.",
-                DeprecationWarning,
-                stacklevel=2,
-            )
-            return request.environ, response  # type: ignore
-
-        return response
+            return response
 
     def get(self, *args: t.Any, **kw: t.Any) -> "TestResponse":
         """Call :meth:`open` with ``method`` set to ``GET``."""
@@ -1271,6 +1278,13 @@ class TestResponse(Response):
     If the test request included large files, or if the application is
     serving a file, call :meth:`close` to close any open files and
     prevent Python showing a ``ResourceWarning``.
+
+    .. versionchanged:: 2.1
+        Removed deprecated behavior for treating the response instance
+        as a tuple.
+
+    .. versionadded:: 2.0
+        Test client methods always return instances of this class.
     """
 
     request: Request
@@ -1283,6 +1297,9 @@ class TestResponse(Response):
     is made with ``follow_redirects`` enabled.
     """
 
+    # Tell Pytest to ignore this, it's not a test class.
+    __test__ = False
+
     def __init__(
         self,
         response: t.Iterable[bytes],
@@ -1297,28 +1314,11 @@ class TestResponse(Response):
         self.history = history
         self._compat_tuple = response, status, headers
 
-    def __iter__(self) -> t.Iterator:
-        warnings.warn(
-            (
-                "The test client no longer returns a tuple, it returns"
-                " a 'TestResponse'. Tuple unpacking is deprecated and"
-                " will be removed in Werkzeug 2.1. Access the"
-                " attributes 'data', 'status', and 'headers' instead."
-            ),
-            DeprecationWarning,
-            stacklevel=2,
-        )
-        return iter(self._compat_tuple)
+    @cached_property
+    def text(self) -> str:
+        """The response data as text. A shortcut for
+        ``response.get_data(as_text=True)``.
 
-    def __getitem__(self, item: int) -> t.Any:
-        warnings.warn(
-            (
-                "The test client no longer returns a tuple, it returns"
-                " a 'TestResponse'. Item indexing is deprecated and"
-                " will be removed in Werkzeug 2.1. Access the"
-                " attributes 'data', 'status', and 'headers' instead."
-            ),
-            DeprecationWarning,
-            stacklevel=2,
-        )
-        return self._compat_tuple[item]
+        .. versionadded:: 2.1
+        """
+        return self.get_data(as_text=True)
